function [output] = stop_foforth(fun, opts, varargin)
% using a gradient reflection (or projection) method to solve
%
%       min             f(X):= E(X)+trace(G'X)
%       s. t.         X'X = I,  where X in R^{n*p}.
%
%  Assume gradient of E(X)=H(X)X, where H(X) is a  n-by-n symmetric matrix.
% ----------------------------------
%  Input:
%                         fun --- a matlab function for f(X)
%                                     call:  [fvalue, gradf] = fun(X, data1, data2);
%                                   fvalue:  function value f(X)
%                                    gradf:  gradient of f(X)
%                                     data:  extra data (can be more)
%
%  Calling syntax:
%                     [output]= FOForth(X0,G,@fun,opts,data1,data2);
%
%                      opts --- option structure with fields:
%                                         G: n-by-p matrix
%                                         X: n-by-p initial point such that X'X=I
%                                    solver:  1(gradient reflection method) 2(projection method) 3(QR retraction)
%                                  stepsize:  0(ABB stepsize) o.w.(fixed stepsize)
%                                      xtol:  stop control for ||X_k - X_{k+1}||/sqr(n)
%                                      gtol:  stop control for ||kkt||/||kkt0||
%                                      ftol:  stop control for ||f_k -f_{k+1}||/(|f_k|+ 1)
%                                     maxit:  max iteration
%                              info_warning:  0(no print) o.w.(print)
%
%  Output:
%                        output --- output information
%                                         X:  solution
%                                       kkt: ||kkt|| (first-order optimality condition)
%                                      fval:  function value of solution
%                                      feaX: ||I-X'X||_F (feasiblity violation)
%                                      xerr: ||X_k - X_{k+1}||/sqr(n)
%                                      iter: total iteration number
%                                     fvals: history of function value
%                                      kkts: history of kkt
%                                   message: convergence message
% --------------------------------------------------------------------
%  Reference:
%   B. Gao, X. Liu, X. Chen and Y. Yuan
%   A new first-order algorithmic framework for optimization problems with
%   orthogonality constraints, SIAM Journal on Optimization, 28 (2018), pp.302--332.
% ----------------------------------
%  Author: Bin Gao, Xin Liu (ICMSEC, AMSS, CAS)
%                 gaobin@lsec.cc.ac.cn
%                 liuxin@lsec.cc.ac.cn
%  Version: 1.0 --- 2016/04/01
%  Version: 1.1 --- 2017/10/16: support general function
%  Version: 1.2 --- 2021/07/26: adaptation to STOP package
% --------------------------------------------------------------------
%% default setting

% warnings
if ~isfield(opts, 'info_warning')
    info_warning = 0;
else
    info_warning = opts.info_warning;
end

% specify the initial point
if isfield(opts, 'X') && (~isempty(opts.X))
    [n, p] = size(opts.X);
    X = opts.X;
elseif isfield(opts, 'dim')
    n = opts.dim(1);
    p = opts.dim(2);
    if n < p
        error('Column size should be no less than the size of row. Please check the dimensions of the opts.dim.') % ????????, n\& p????????????????
    end
    [X, ~] = qr(randn(n, p), 0);
else
    error('No initial point, and the size of initial point is not specified.')
end

% specify G
if isfield(opts, 'G') && (~isempty(opts.G))
    G = opts.G;
else
    G = zeros(n,p);
    warning('G is not specified. Running with G=0 by defauly!')
end

% Specify the options for stepsize
if isfield(opts, 'stepsize')
    if ~isfield(opts.stepsize, 'type')
        stepsize_type = 0;
        if info_warning
            warning('No input stepsize type, choose alternating BB stepsize.')
        end
    else
        if strcmp(opts.stepsize.type, 'ABB')
            stepsize_type = 0;
        elseif strcmp(opts.stepsize.type, 'BB1')
            stepsize_type = 1;
        elseif strcmp(opts.stepsize.type, 'BB2')
            stepsize_type = 2;
        elseif strcmp(opts.stepsize.type, 'SD') % initial stepsize by difference
            stepsize_type = 3;
        else
            stepsize_type = 0;
            if info_warning
                warning('Invaild stepsize type, choose ABB instead.')
            end
        end
    end
    
    if ~isfield(opts.stepsize, 'max_stepsize')
        max_stepsize = 1000;
    else
        max_stepsize = opts.stepsize.max_stepsize;
    end
    
    if ~isfield(opts.stepsize, 'min_stepsize')
        min_stepsize = 0;
    else
        min_stepsize = opts.stepsize.min_stepsize;
    end
    
    if ~isfield(opts.stepsize, 'init_stepsize')
        init_stepsize = 1e-3;
    else
        init_stepsize = opts.stepsize.init_stepsize;
    end
else
    if info_warning
        warning('No input stepsize.')
    end
    stepsize_type = 0;
    max_stepsize = 1000;
    min_stepsize = 0;
    init_stepsize = 1;
end

% Options for line search
if ~isfield(opts, 'linesearch')
    linesearch = 0;
else
    linesearch = opts.linesearch;
end

%% Maximum iterations
if ~isfield(opts, 'maxit')
    maxit = 3000;
else
    maxit = opts.maxit;
    if maxit < 0 || maxit > 10000
        maxit = 3000;
    end
end

% Options for tolerence
if ~isfield(opts, 'gtol') % add gtol
    gtol = 1e-3;
else
    gtol = opts.gtol;
    if gtol < 0 || gtol > 1
        gtol = 1e-3;
        if info_warning
            warning('Invaild value of gtol, choose 1e-3 instead.')
        end
    end
    
end
if ~isfield(opts, 'xtol') % add xtol
    xtol = 1e-6;
else
    xtol = opts.xtol;
    if xtol < 0 || xtol > 1
        xtol = 1e-6;
        if info_warning
            warning('Invaild value of xtol, choose 1e-6 instead.')
        end
    end
    
end

% Options for printed infomation
if ~isfield(opts, 'info')
    local_info = 2;
else
    local_info = opts.info;
end

% Specify the solver
if isfield(opts, 'solver')
    if strcmp(opts.solver, 'GR')
        solver = 1;
    elseif strcmp(opts.solver, 'GP')
        solver = 2;
    elseif strcmp(opts.solver, 'QR')
        solver = 3;
    else
        solver = 2;
        if info_warning
            warning('Invaild solver type, choose GP instead.')
        end
    end
else
    if info_warning
        warning('No input solver, choose GP by default.')
    end
    solver = 2;
end

%% ---------------------------------------------------------------
global Ip
Ip = eye(p);

%% ---------------------------------------------------------------
% Initialization
iter = 0; 
% ensure X is orthogonal
if norm(X'*X-Ip,'fro')>1e-13; [X,~] = qr(X,0); end
% evaluate function and gradient info.
[funX, F] = feval(fun, X , varargin{:});
[PF,kkt0,feaX] = getPG(X,F);
% save history

violation_KKT = [kkt0];
violation_Constraint = [feaX];
fval = [funX];

time_record = [0];
t_start = tic;
line_search_record = [0];

% initial stepsize
% if ~isfield(opts.stepsize, 'init_stepsize')
init_stepsize = max(0.1,min(0.01*kkt0,1));
% end

tau = init_stepsize;

% initial solver
switch solver
    case 1; mainsolver = @gradient_reflection;
    case 2; mainsolver = @projection;
    case 3; mainsolver = @QRretraction;
end

% info
if local_info ~= 0
    switch solver
        case 1
            fprintf('------------------ FOForth with gradient reflection start ------------------\n');
        case 2
            fprintf('------------------ FOForth with projection start ------------------\n');
        case 3
            fprintf('------------------ Riemannian Opt with QR retraction start ------------------\n');
    end
    fprintf('%4s | %15s | %10s | %10s | %8s | %8s\n', 'Iter ', 'F(X) ', 'KKT ', 'Xerr ', 'Feasi ', 'tau');
    fprintf('%d \t %f \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',iter, funX, kkt0, 0, feaX, tau);
end

%% ---------------------------------------------------------------
% main iteration
for iter = 1:maxit
    Xk = X;     Fk = F;    PFk = PF;   funXk = funX;
    
    % ---------- PART I: gradient step ----------
    %  Riemmanian or Euclidean gradient
    if solver == 3
        Grad = PFk;
    else
        Grad = Fk;
    end
    
    if linesearch == 0
        % gradient step
        V = Xk - tau * Grad;
        X = mainsolver(Xk,V);
        ls_iter = 0;
    else
        local_fval_grad = @(ZZ) feval(fun, ZZ, varargin{:});
        if stepsize_type ~= 3
            [X, ls_iter] = FOForth_lineasarch(Xk, Grad, tau, local_fval_grad, mainsolver, output.fvals);
        else
            [X, ls_iter] = FOForth_lineasarch_sd(Xk, Grad, tau, local_fval_grad, mainsolver, output.fvals);
        end
    end
    line_search_record = [line_search_record, ls_iter];
    
    
    % ---------- PART II: symmetrization step ----------
    if norm(X'*X-Ip,'fro')>1e-12; [X,~] = qr(X,0); end
    if solver ~= 3
        if G ~= 0
            [tu,~,tv] = svd(X'*G,0);
            X = -X*(tu*tv');
        end
    end
    
    % ------------ evaluate error ------------
    [funX, F] = feval(fun, X , varargin{:});
    [PF,kkt,feaX] = getPG(X,F);
    
    violation_KKT = [violation_KKT, kkt];
    violation_Constraint = [violation_Constraint, feaX];
    time_record = [time_record, toc(t_start)];
    fval = [fval, funX];
       
    xerr = norm(Xk - X,'fro')/sqrt(n);
    ferr = abs(funXk - funX)/(abs(funXk)+1);
    
    % info
    if local_info ~= 0 && (mod(iter,20) == 0 )
        fprintf('%d \t %f \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',iter, funX, kkt, xerr, feaX, tau);
    end
    
    % ------------ update ABB stepsize ------------
    switch stepsize_type
        case 0
            Sk = X - Xk;
            Vk = PF - PFk;     %Vk = F - Fk;
            SV = sum(sum(Sk.*Vk));
            if mod(iter+1,2) == 0
                tau = abs(SV)/sum(sum(Vk.*Vk)); % SBB for odd
            else
                tau = sum(sum(Sk.*Sk))/abs(SV);  % LBB for even
            end
        case 1
            Sk = X - Xk;
            Vk = PF - PFk;     %Vk = F - Fk;
            SV = sum(sum(Sk.*Vk));
            tau = abs(SV)/sum(sum(Vk.*Vk)); % SBB
            
        case 2
            Sk = X - Xk;
            Vk = PF - PFk;     %Vk = F - Fk;
            SV = sum(sum(Sk.*Vk));
            tau = sum(sum(Sk.*Sk))/abs(SV);  % LBB
        case 3
            tau = abs(funXk - funX)/norm(Grad,'fro')^2;
    end
    tau = max(min_stepsize, min(abs(tau), max_stepsize));
    
    
    % ------------------ stop criteria --------------------
    if kkt/kkt0 < gtol
    %     if kkt/kkt0 < gtol || (xerr < xtol || ferr < ftol)
    %     if kkt/kkt0 < gtol || (xerr < xtol && ferr < ftol)
    %     if (kkt/kkt0 < gtol && xerr < xtol) || ferr < ftol
    %     if (kkt/kkt0 < gtol && ferr < ftol) || xerr < xtol
%     if kkt/kkt0 < gtol || (xerr < xtol && ferr < ftol)
        output.message = 'converge';
        break;
    end
end

if iter >= maxit
    output.message = 'exceed max iteration';
end

output.fea = feaX;
output.fval = funX;
output.iter = iter;
output.xerr = xerr;
output.time = time_record(end);
output.kkt = kkt;

output.kkts = violation_KKT;
output.fvals = fval;
output.feas = violation_Constraint;
output.times = time_record;

output.line_search_iters = line_search_record;


if local_info ~= 0
    fprintf('%s at...\n',output.message);
    fprintf('%d \t %f \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',iter, funX, kkt, xerr, feaX, tau);
    fprintf('------------------------------------------------------------------------\n');
end

%% ---------------------------------------------------------------
% nest-function
% gradient reflection step
    function X = gradient_reflection(X,V)
        VV = V'*V;
        VX = V'*X;
        TVX = VV\VX;
        X = -X + V*(2*TVX);
    end

% projection step
    function X = projection(~,V)
        % approach 1
        VV = V'*V;
        [Q,D] = eig(VV);
        DD = sqrt(D)\Ip;
        X = V*(Q*DD*Q');
        
        % approach 2
        %  [UX,~,VX] = svd(V,0);
        %   X = UX*VX';
    end

% QR retraction
    function X = QRretraction(~,V)
        % approach 1
        VV = V'*V;
        L = chol(VV,'lower');
        X = V*inv(L)';
        
        % approach 2
        %  [X,~] = qr(V,0);
    end

% get projected gradient and its norm
    function [PF,normPF,feaX] = getPG(X,F)
        PF = F - X*(F'*X);
        normPF = norm(PF,'fro');
        feaX = norm(X'*X-Ip,'fro');
    end

end