function [output] = stop_pcal(fun, opts, varargin)
% -----------------------------------------------------------------------
% Using a Parallel Column-wise Block Minimization
% (for Proximal Linearized Augmented Lagrangian )
% to solve
%
%       min  f(X),  s. t.  X'MX = I,  where X\in R^{n,p}
%
%  where M is a symmetric positive definite matrix.
% ----------------------------------
%  Input:
%                         X --- n-by-p initial matrix such that X'*M*X = I
%                     fun --- a matlab function for f(X)
%                                    call: [funX,F] = fun(X,data1,data2);
%                                    funX: function value f(X)
%                                           F:  gradient of f(X)
%                                    data: extra data (can be more)
%          varargin --- data1, data2

%
%  Calling syntax:
%      If M is an identity matrix, i.e., X'X = I.
%       opts.M=[]; otherwise, opts.M=M;
%      [X, out]= PCAL(X0, @fun, opts, data1, data2);
%
%                     opts --- option structure with fields:
%                                               M:  n-by-n symmetric positive definite matrix
%                                        xtol:  stop control for ||X_k - X_{k+1}||/sqrt(n)
%                                        gtol:  stop control for ||kkt||/||kkt0||
%                                        ftol:  stop control for |f_k - f_{k+1}|/(|f_k|+1)
%                               stepsize:  0(ABB stepsize) o.w.(fixed stepsize)
%                           penalparam:  penalty factor
%                                                     This solver is sensitive to penalty parameter. Users need
%                                                      to tune different values with different problem settings.
%                                    solver:  1 (PCAL)  2 (PLAM)   3 (PCAL-S, simplified multiplier)
%                                postorth:  1 (post-procedure)  0 (no post-procedure)
%                                      maxit:  max iteration
%                                        info:   0(no print) o.w.(print)
%
%  Output:
%                            X --- solution
%                        output --- output information
%                                        kkt: ||kkt|| (first-order optimality condition)
%                                      fval:  function value of solution
%                                      feaX: ||I-X'X||_F (feasiblity violation)
%                                      xerr: ||X_k - X_{k+1}||/sqr(n)
%                                      iter: total iteration number
%                                    fvals: history of function value
%                                    feaXs: history of feasibility
%                                      kkts: history of kkt
%                                message: convergence message
% --------------------------------------------------------------------
%  Reference:
%  B. Gao, X. Liu, and Y.-x. Yuan, Parallelizable algorithms for optimization
%  problems with orthogonality constraints, SIAM Journal on Scientific Computing, 41-3 (2019), A1949???A1983.
% ----------------------------------
%  Author: Bin Gao, Xin Liu (ICMSEC, AMSS, CAS)
%                 gaobin@lsec.cc.ac.cn
%                 liuxin@lsec.cc.ac.cn
%
%  Version: 1.0 --- 2016/12/22
%  Version: 1.1 --- 2018/02/20: support general function
%  Version: 1.2 --- 2019/01/29: support X'MX=I
%  Version: 1.3 --- 2021/07/26: adaptation to STOP package
%---------------------------------------------------------------
%% default setting

% warnings
if ~isfield(opts, 'info_warning')
    info_warning = 0;
else
    info_warning = opts.info_warning;
end

% specify the initial point
if isfield(opts, 'X') && (~isempty(opts.X))
    [n, p] = size(opts.X);
    X = opts.X;
elseif isfield(opts, 'dim')
    n = opts.dim(1);
    p = opts.dim(2);
    if n < p
        error('Column size should be no less than the size of row. Please check the dimensions of the opts.dim.') % ????????, n\& p????????????????
    end
    [X, ~] = qr(randn(n, p), 0);
else
    error('No initial point, and the size of initial point is not specified.')
end

if isfield(opts, 'M') && ~isempty(opts.M)
    if ~issymmetric(opts.M);error('input M is not a symmetric matrix');end
    general_flag = 1;
else
    general_flag = 0;
end

% Specify the options for stepsize
if isfield(opts, 'stepsize')
    if ~isfield(opts.stepsize, 'type')
        stepsize_type = 0;
        if info_warning
            warning('No input stepsize type, choose alternating BB stepsize.')
        end
    else
        if strcmp(opts.stepsize.type, 'ABB')
            stepsize_type = 0;
        elseif strcmp(opts.stepsize.type, 'BB1')
            stepsize_type = 1;
        elseif strcmp(opts.stepsize.type, 'BB2')
            stepsize_type = 2;
        elseif strcmp(opts.stepsize.type, 'SD') % initial stepsize by difference
            stepsize_type = 3;
        else
            stepsize_type = 0;
            if info_warning
                warning('Invaild stepsize type, choose ABB instead.')
            end
        end
    end
    
    if ~isfield(opts.stepsize, 'max_stepsize')
        max_stepsize = 1000;
    else
        max_stepsize = opts.stepsize.max_stepsize;
    end
    
    if ~isfield(opts.stepsize, 'min_stepsize')
        min_stepsize = 0;
    else
        min_stepsize = opts.stepsize.min_stepsize;
    end
    
    if ~isfield(opts.stepsize, 'init_stepsize')
        init_stepsize = 1e-3;
    else
        init_stepsize = opts.stepsize.init_stepsize;
    end
else
    if info_warning
        warning('No input stepsize.')
    end
    stepsize_type = 0;
    max_stepsize = 1000;
    min_stepsize = 0;
    init_stepsize = 1;
end

% Options for line search
if ~isfield(opts, 'linesearch')
    linesearch = 0;
else
    linesearch = opts.linesearch;
end

%% Maximum iterations
if ~isfield(opts, 'maxit')
    maxit = 3000;
else
    maxit = opts.maxit;
    if maxit < 0 || maxit > 10000
        maxit = 3000;
    end
end

% Options for tolerence
if ~isfield(opts, 'gtol') % add gtol
    gtol = 1e-3;
else
    gtol = opts.gtol;
    if gtol < 0 || gtol > 1
        gtol = 1e-3;
        if info_warning
            warning('Invaild value of gtol, choose 1e-3 instead.')
        end
    end
    
end
if ~isfield(opts, 'xtol') % add xtol
    xtol = 1e-6;
else
    xtol = opts.xtol;
    if xtol < 0 || xtol > 1
        xtol = 1e-6;
        if info_warning
            warning('Invaild value of xtol, choose 1e-6 instead.')
        end
    end
    
end

% Options for printed infomation
if ~isfield(opts, 'info')
    local_info = 2;
else
    local_info = opts.info;
end


if isfield(opts, 'proxparam')
    if opts.proxparam < 0
        opts.proxparam = 0;
    end
else
    opts.proxparam = 0;
end

% 1.PCAL  2.PLAM  3.PCAL-S
if isfield(opts, 'solver')
    if all(opts.solver ~= 1:2)
        opts.solver = 1;
    end
else
    opts.solver = 1;
end

if isfield(opts, 'postorth')
    if all(opts.postorth ~= [0 1])
        opts.postorth = 1;
    end
else
    opts.postorth = 1;
end

%% ---------------------------------------------------------------
% copy parameters
solver = opts.solver;
postorth = opts.postorth;


global Ip
Ip = eye(p);

%% ---------------------------------------------------------------
% Initialization
iter = 0; iter_final=0;
% evaluate function and gradient info.
[funX,G] = feval(fun, X , varargin{:});
if general_flag; M = opts.M; MX = M*X; else; MX = X;end
[PL,kktval0,feaX] = getPG(X,G,MX,0);
% save history
violation_KKT = [kktval0];
violation_Constraint = [feaX];
fval = [funX];

time_record = [0];
t_start = tic;
line_search_record = [0];

% Specify the penalty parameter according to ||\nabla f(X_0)||_F
if ~isfield(opts, 'penalparam')
    penalparam = max(min(1*norm(G,'fro'), 1e10), 1e-10); % set lower-bound for beta
else
    penalparam = opts.penalparam;  % warning for too large/small beta
    if info_warning
        if penalparam > 1e10
            warning('The penalty parameter is too large. Please scale the problem or specify opts.penalparam.');
        elseif penalparam < 1e-10
            warning('The penalty parameter is too small. Please scale the problem or specify opts.penalparam.');
        end
    end
end

% initial stepsize
proxparam = 1/max(0.1,min(0.01*norm(PL,'fro'),1));


% info
if local_info ~= 0
    switch solver
        case 1
            fprintf('------------------ PCAL start ------------------\n');
        case 2
            fprintf('------------------ PLAM start ------------------\n');
        case 3
            fprintf('------------------ PCAL-S start ------------------\n');
    end
    fprintf('%4s | %15s | %10s | %10s | %8s | %8s\n', 'Iter ', 'F(X) ', 'KKT ', 'Xerr ', 'Feasi ', 'tau');
    fprintf('%d \t %f \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',iter, funX, kktval0, 0, feaX, 1/proxparam);
end

%% main iteration
for iter = 1: maxit
    Xk = X; PLk = PL;   funXk = funX;
    
    
    if linesearch == 0
        % one gradient step for PCAL, PLAM
        X = Xk - PLk/proxparam;
        if general_flag
            MX = M*X;
        else
            MX = X;
        end
        ls_iter = 0;
    else
        local_fval_grad = @(ZZ) feval(fun, ZZ, varargin{:});
        mainsolver = 0;
        if stepsize_type ~= 3
            [X, ls_iter] = PCAL_lineasarch(Xk, PLk, 1/proxparam, local_fval_grad, mainsolver, output.fvals);
        else
            [X, ls_iter] = PCAL_lineasarch_sd(Xk, PLk, 1/proxparam, local_fval_grad, mainsolver, output.fvals);
        end
    end
    line_search_record = [line_search_record, ls_iter];
    
    
    % ------------ Only for PCAL ------------
    if any(solver == [1 3])
        % -------- solve subproblems -----------
        v = sqrt(sum(X.*MX)); % column square root
        X = X./v;
        if general_flag
            MX = MX./v;
        else
            MX = X;
        end
    end
    
    % ------------ evaluate error ------------
    [funX,G] = feval(fun, X , varargin{:});
    [PL,kktval,feaX] = getPG(X,G,MX,penalparam);
    
    % ------------ save history ------------
    violation_KKT = [violation_KKT, kktval];
    violation_Constraint = [violation_Constraint, feaX];
    time_record = [time_record, toc(t_start)];
    fval = [fval, funX];
    xerr = norm(Xk - X,'fro')/sqrt(n);
    ferr = abs(funXk - funX)/(abs(funXk)+1);
    % info
    if local_info ~= 0 && (mod(iter,20) == 0 )
        fprintf('%d \t %f \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',iter, funX, kktval, xerr, feaX, 1/proxparam);
    end
    
    % --------- parameter: proxparam -------------
    switch stepsize_type
        case 0
            Sk = X-Xk;
            Vk = PL-PLk;    % Vk = G-Gk;
            SV = sum(sum(Sk.*Vk));
            if mod(iter+1,2) == 0
                proxparam = sum(sum(Vk.*Vk))/abs(SV); % SBB for odd
            else
                proxparam = abs(SV)/sum(sum(Sk.*Sk)); % LBB for even
            end
        case 1
            Sk = X-Xk;
            Vk = PL-PLk;    % Vk = G-Gk;
            SV = sum(sum(Sk.*Vk));
            proxparam = sum(sum(Vk.*Vk))/abs(SV); % SBB
        case 2
            Sk = X-Xk;
            Vk = PL-PLk;    % Vk = G-Gk;
            SV = sum(sum(Sk.*Vk));
            proxparam = abs(SV)/sum(sum(Sk.*Sk)); % LBB
        case 3
            proxparam = norm(PLk,'fro')^2/abs(funXk - funX);
    end
    proxparam = max(min_stepsize,min(proxparam,max_stepsize));
    
    % ------------------ stop criteria --------------------
    if (kktval/kktval0 < gtol)
        %     if (kktval/kktval0 < gtol && feaX < 1.e-12)
        %     if (kktval/kktval0 < gtol)  || (xerr < xtol && ferr < ftol)
        output.message = 'converge';
        iter_final = iter;
        % ----- post-procedure: orthogonal step -----
        if postorth == 1
            if feaX > 1e-13
                % projection to orthogonality constraints
                if general_flag
                    R = chol(M); RX = R*X;
                    [U,~,V] = svd(RX,0); B = U*V';  % projection
                    opts_orth.UT = true; X = linsolve(R,B,opts_orth); % solve RX = B
                else
                    [U,~,V] = svd(X,0);   X = U*V';   % projection
                end
                
                iter_final = iter+1;
                [funX,G] = feval(fun, X , varargin{:});
                if general_flag; MX = M*X; else; MX = X;end
                [~,kktval,feaX] = getPG(X,G,MX,penalparam);
                violation_KKT = [violation_KKT, kktval];
                violation_Constraint = [violation_Constraint, feaX];
                time_record = [time_record, toc(t_start)];
                fval = [fval, funX];
                xerr = norm(Xk - X,'fro')/sqrt(n);
                ferr = abs(funXk - funX)/(abs(funXk)+1);
            end
        end
        break;
    end
    
    if iter >= maxit
        iter_final = maxit;
        output.message = 'exceed max iteration';
        break;
    end
    
    % ------------- adaptive parameter: penalparam ------------
    %     if  feaX > 1.e-13
    %         penalparam = 2*penalparam;
    %     end
    % ------------------------------------------
    
end

output.fea = feaX;
output.fval = funX;
output.iter = iter_final;
output.xerr = xerr;
output.kkt  = kktval;
output.time = time_record(end);

output.kkts = violation_KKT;
output.fvals = fval;
output.feas = violation_Constraint;
output.times = time_record;

if local_info ~= 0
    if iter_final > iter
        fprintf('%s at %d-th and post-procedure is calling...\n',output.message,iter_final-1);
    else
        fprintf('%s at ... (without post-procedure)\n',output.message);
    end
    fprintf('%d \t %f \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',iter_final, funX, kktval, xerr, feaX, 1/proxparam);
    fprintf('------------------------------------------------------------------------\n');
end

%% ---------------------------------------------------------------
% nest-function
% get Lagrangian gradient: PL
% ||G-MXG'X||: kktval
% ||I-X'MX||: feaX
    function [PL,kktval,feaX] = getPG(X,G,MX,penalparam)
        GX = G'*X;  % grad'*X
        GXsym = 0.5*(GX+GX');
        XGXsym = MX*GXsym;
        XX = X'*MX;
        FeaX = XX-Ip;  % X'X-I
        
        kkt = G - XGXsym;
        kktval = norm(kkt,'fro');
        feaX = norm(FeaX,'fro');
        
        if penalparam ~= 0
            penalFeaX = penalparam*FeaX; % beta*(X'MX-I)
        end
        
        % -------- Lambda & PL --------
        switch solver
            case 1   % PCAL
                %  PG = G - MX*GXsym; % kkt  (G-MXG'X)
                %  Lambda = GXsym + diag(diag(MX'*(PG-penalPX)));
                if penalparam ~= 0
                    d = diag(GX'-XX*GXsym+XX*penalFeaX);
                    % Lambda = GXsym + diag(d);
                    PL = kkt - MX.*d' + MX*penalFeaX;
                else
                    d = diag(GX'-XX*GXsym);
                    % Lambda = GXsym + diag(d);
                    PL = kkt - MX.*d'; % Lambda = XGXsym + XD;
                end
            otherwise  % PLAM, PCAL-S
                % Lambda = GXsym;
                if penalparam ~= 0
                    PL = kkt + MX*penalFeaX;
                else
                    PL = kkt;
                end
        end
        % -------------------
    end
% -------------------------------------------------------
end
