function [output] = stop_pencs(fun,opts,varargin)
% Second-order method for smooth optimization over the Stiefel manifold based on PenC model (PenCS)
% function [output] = stop_pencs(fun,opts)
% function [output] = stop_pencs(fun,opts,varargin)

% The descriptions for stop_pencs solver could also be found at https://stmopt.gitee.io/algorithm_description/PenCS_code.html
% This solver is a part of stmopt package (https://stmopt.gitee.io/)

% The stop_pencs solver starts at opts.X if it is provided (otherwise, generates opts.X whose size is specified in
% opts.dim on the Stiefel manifold). 
% The fun in stop_pencs solver is called by
% [fval, grad] = fun(X) or [fval, grad] = fun(X, varargin)

% The opts in stop_pencs solver is a matlab structure for overwritting the default values. 
% In opts structure, either opts.X or opts.dim should be specified. 
% Any other option have a default value and is hence optional. To assign the value to the opts structure,
% one needs to pass an options structure with a field options.optionname.

% * * opts.X
%     * The initial point for the solver. opts.X should be a matrix with $n$ rows and $p$ columns, where $n\geq p$. We recommend to choose a feasible opts.X in practice. 
%     * When opts.X is empty, the option opts.dim should be specified, and the solver will randomly generate a feasible initial point by [X, ~] = qr(randn(n,p),0); otherwise, the solver will raise an error. 
%   * opts.proj
%     * Projection to a set that contains Stiefel manifold
%     * The default value is set as opts.proj = @(X)  X * min(1.0001*(sqrt(p)/norm(X,'fro')), 1) ;. This is equivalent to project $X$ into a ball with radius $1.001 \sqrt{p}$.
%   * opts.stepsize
%     *  Stepsize arguments for PenCS
%     *  opts.stepsize.type = 0:  
%        *  opts.stepsize.type = 'ABB': Alternating BB stepsize
%        *  opts.stepsize.type = 'BB1': BB1 stepsize
%        *  opts.stepsize.type = 'BB2': BB2 stepsize
%        *  opts.stepsize.type = 'FIX': fixed stepsize
%     *  opts.stepsize.max_stepsize = 1000: Upper-bound for stepsize
%     *  opts.stepsize.min_stepsize = 0: Lower-bound for stepsize
%     *  opts.stepsize.init_stepsize: Initial stepsize
%   * opts.gtol = 1e-10
%     * Tolerance for the kkt condition
%     * PenCS terminates when $\frac{||\nabla f(X_k) - X_k \Lambda(X_k) ||_F}{||\nabla f(X_0) - X_0 \Lambda(X_0) ||_F} \leq gtol$,
%       Here $\Lambda(X) := 0.5*(X^T\nabla f(X) + \nabla f(X)^T X)$.
%   * opts.maxit = 1000
%     * Maximum iteration
%     * PenCS terminates when the iteration exceeds maxit
%   * opts.penalparam
%     * Penalty parameter 
%     * The penalty parameter is important to the performance of PenCS. A small penalty parameter leads to diverge while a large one results in slow convergence rate. 
%     * The default value for opts.penalparam is chosen as $ ||\nabla f(X_0)||_F$.
%   * opts.safe_proj
%     * Option for safeguards
%     * safe_proj = 1 is a function that pulls $X_k$ close to Stiefel manifold if it is a little far away from Stiefel manifold. It improves the robustness of PenCS, but takes additional computational costs. 
%   * opts.postorth = 1
%     * Option for orthonormalization as postprocess:
%     * The postprocess improves the feasibility but requires additional computational costs. We recommand to set  opts.postorth = 1 when high precision in feasibility is required. 
%   * opts.info = 0
%     * Option for display during the execution
%     * No print or display when opts.info = 0
%   * opts.info_warning = 0
%     * Options for displaying warnings in the solver. 




% All the outputs of stop_pencs is wrapped in the output structure. 

% out: output information

% * out.X
%   * Final solution
% * out.iter
%   * Total iterations taken by PenCS
% * out.fval
%   * Final function value
% * out.kkt
%   * Final substationarity evaluated by $||\nabla f(X_k) - X_k \Lambda(X_k)||_F$
% * out.fea
%   * Final feasibility evaluated by $||X_k^\top X_k - I_p||_F$ 
% * out.fvals
%   * Records of function values in each iteration
% * out.kkts
%   * Record of the substationarities in each iteration
% * out.feas
%   * Record of the feasibility in each iteration
% * out.times
%   * Records of the CPU time in each iteration. 
% * out.subiter
%   * Records of the iterations taken by CG solver in each iteration. 

% The implementation of this solver is based on the following paper:

% N. Xiao, X. Liu and Y. Yuan, A Class of Smooth Exact Penalty Function Methods for Optimization Problems with Orthogonality Constraints, Optimization Methods and Software 
% (DOI:10.1080/10556788.2020.1852236). 
% https://www.tandfonline.com/doi/abs/10.1080/10556788.2020.1852236?journalCode=goms20
% 
if ~isfield(opts, 'info_warning')
    info_warning = 0;
else
    info_warning = opts.info_warning;
end
if isfield(opts, 'X') && (~isempty(opts.X))
    [n,p] = size(opts.X);
    X = opts.X;
elseif isfield(opts, 'dim')
    n = opts.dim(1);
    p = opts.dim(2); % 检查dim的size
    if n < p
        error('Column size should be no less than the size of row. Please check the dimensions of the opts.dim.') % 否定句, n\& p的具体含义
    end
    [X, ~] = qr(randn(n,p),0);
else
    error('No initial point, and the size of initial point is not specified')
end

Hessf = opts.Hessf;

if isfield(opts, 'X') && (~isempty(opts.X))
    [n,p] = size(opts.X);
    X = opts.X;
elseif isfield(opts, 'dim')
    n = opts.dim(1);
    p = opts.dim(2); % 检查dim的size
    if n < p
        error('Column size should be no less than the size of row. Please check the dimensions of the opts.dim.') % 否定句, n\& p的具体含义
    end
    [X, ~] = qr(randn(n,p),0);
else
    error('No initial point, and the size of initial point is not specified')
end


if ~isfield(opts, 'maxit')
    maxlength = 3000;
else
    maxlength = opts.maxit;
end


[fvalue, gradf] = feval(fun,X,varargin{:});
s_upper = norm(gradf,'fro');

if ~isfield(opts, 'penalparam')
    beta = max(min(1*s_upper/p, 1e10), 1e-10); % set lower-bound for beta
else
    beta = opts.penalparam;  % warning for too large/small beta
    if info_warning
        if beta > 1e10
            warning('The penalty parameter is too large. Please scale the problem or specify opts.penalparam.');
        elseif beta < 1e-10
            warning('The penalty parameter is too small. Please scale the problem or specify opts.penalparam.');
        end
    end
end

Lambda = X'*gradf;
Lambda = 0.5*(Lambda+Lambda');
XX = X'*X;
Grad = gradf - X*(Lambda  - beta * ( XX- eye(p))  )  ; % approximated gradient
Grad_tmp = gradf - X * Lambda;


if ~isfield(opts,'proj')
    proj = @(X)  X * min(1.0001*(sqrt(p)/norm(X,'fro')), 1) ;
else
    proj = opts.proj;
end


if ~isfield(opts,'gtol') % add gtol
    tol_all = 1e-10;
else
    tol_all = opts.gtol;
end



if ~isfield(opts,'xtol') % add xtol
    xtol = 1e-15;
else
    xtol = opts.xtol;
end

%% Options for post-process
if ~isfield(opts,'postorth')
    postprocess = 1;
else
    postprocess = opts.postorth;
end



violation_KKT = zeros(1,maxlength);
violation_Constraint = zeros(1,maxlength);
fval = zeros(1,maxlength);
CPU_iter = zeros(1,maxlength);

para.beta = beta;

[~,gradf] = feval(fun,X,varargin{:});
Lambda = X'*gradf;
Lambda = 0.5*(Lambda+Lambda');
XX = X'*X;
HXXXI = feval(Hessf,X,X*(XX-eye(p)),varargin{:});
grad = gradf*(1.5*eye(p) - 0.5*XX) - X*(Lambda - beta*(XX - eye(p))) - 0.5*HXXXI;
num_iter = zeros(maxlength,1);
radius_tr_coeff = 1;
for jj = 1:1:maxlength
    para.iter = jj;
    para.tol = tol_all;
    para.beta = beta;
    para.XX = XX;
    para.Lambda = Lambda;
    para.gradf = gradf;
    norm_grad = norm(grad,'fro')/sqrt(p);
    num_pre = 0;
    if jj <= 2
        radius_tr = min(10*norm(grad,'fro')^1,1e-1*norm(grad,'fro'));
    else
        radius_tr = min(10*norm(grad,'fro')^1,1* 1e-2*norm(grad,'fro'));
    end
    
    
    %norm_grad = 1;
    grad_reg = grad/norm_grad;
    if  jj <= num_pre && norm_grad > 1e-9
        [D,num_subiter] = CGN(@work_Hessh,fun,Hessf,X,grad_reg,para,(1/norm_grad)*max(0.9*tol_all,radius_tr),varargin{:});
    else
        [D,num_subiter] = CGN(@work_Hessh,fun,Hessf,X,grad_reg,para,(1/norm_grad)*max(0.99*tol_all,radius_tr),varargin{:});
    end
    D = D*norm_grad;
    X_p = X-D;
    X = X_p;
    num_iter(jj) = num_subiter;
    
    if norm(grad,'fro') > 1e-5
        [X,~] = qr(X,0);
    end
    r_C = 10*sqrt(p);
    normX = norm(X,'fro');
    if normX > r_C 
        X = X*(r_C/normX);
    end
    
    X = 1.5*X - 0.5*X*(X'*X);
    
    [fvalue,gradf] = feval(fun,X,varargin{:});
    fval(jj) = fvalue;
    
    
    Lambda = X'*gradf;
    Lambda = 0.5*(Lambda+Lambda');
    XX = X'*X;
    gradL = gradf - X*Lambda;
    HXXXI = feval(Hessf,X,X*(XX-eye(p)),varargin{:});
    grad = gradf*(1.5*eye(p) - 0.5*XX) - X*(Lambda - beta*(XX - eye(p))) - 0.5*HXXXI;
    
    para.XX = XX;
    para.Lambda = Lambda;
    para.gradf = gradf;
    
    violation_KKT(jj) = norm(gradL,'fro');
    violation_Constraint(jj) = norm(XX- eye(p),'fro');
    CPU_iter(jj) = toc;
    
    beta = 1.05*norm(gradf,2);
    fprintf('PenCS|  f: %8.6e, nrmG: %2.1e,  cpu: %8.4f, OutIter: %3d, InnerIter: %4d, nfe: %4d,\n',...
                    fvalue,violation_KKT(jj), CPU_iter(jj), jj, num_subiter, 000);
    if violation_KKT(jj) < tol_all
        [X,~] = qr(X,0);
        violation_Constraint(jj) = norm(XX- eye(p),'fro');
        break;
    end
    
    if violation_KKT(jj) > 1e12
        jj = maxlength;
        break;
    end
    
end




output.kkts = violation_KKT(1:jj);
output.X = X;
output.iter = jj;
output.fval = fval(1:jj);
output.feas = violation_Constraint(1:jj);
output.subiter = sum(num_iter);
output.times = CPU_iter(1:jj);
end


function [X,num_iter] = CGN(work_Hessh,fun,Hessf,Xk,b,para,tol,varargin)
% Conjugate Gradient Method.

[n,p] = size(b);
    




%     if nargin<6
%         tol=1e-15;
%     end
    %AX = A*Xk;
    
    
    
    normb = norm(b,'fro');
    X = zeros(n,p);
    grad = feval(work_Hessh,fun,Hessf,Xk,X,para,varargin{:});
    r = b - grad;
    
    r_init_norm = norm(r,'fro');
    if r_init_norm < tol
        num_iter = 1;
        return
    end
    y = r;
    rr = norm(r,'fro')^2;
    Ay = feval(work_Hessh,fun,Hessf,Xk,y,para,varargin{:});
    %[Ay,~] = feval(fun,y,A,Xk,AX,beta);
    z = Ay;
    s = sum(sum(y.*z)); % y^TAy
    t = rr/s; 
    X = X + t*y;
  
    
    for k = 1:(max(4000, 4*p))
       rr_p = rr;
       r = r - t*Ay;
       rr = norm(r,'fro')^2;
       %normX = norm(X,'fro');
       if( sqrt(rr) < tol) 
           %disp(k)
           num_iter = k;
           return;
       end
       B = rr/rr_p;
       y = r + B*y;
       Ay = feval(work_Hessh,fun,Hessf,Xk,y,para,varargin{:});
       %[Ay,~] = feval(fun,y,A,Xk,AX,beta);
       %z = Ay;
       s = sum(sum(y.*Ay));
       t = rr/s;
       X = X + t*y;
    end
    num_iter = k;
    
end
 

function [output]  = work_Hessh(fun,Hessf,X,D,para,varargin)
% Thie function computes the Hessian of h(X) based on the gradient and
% Hessain of f(X) at point X
% The optput is \nabla^2 h(X)[D]
[n,p] = size(X);

if ~isfield(para,'XX')
    XX = X'*X;
else
    XX = para.XX;
end

if ~isfield(para,'beta')
    beta = 0.1;
else
    beta = para.beta;
end


if ~isfield(para,'gradf')
    [~,grad] = feval(fun,X,varargin{:});
else
    grad = para.gradf;
end

if ~isfield(para,'Lambda')
    Lambda = X'*grad;
    Lambda = 0.5*(Lambda+Lambda');
else
    Lambda = para.Lambda;
end


HD = feval(Hessf,X,D,varargin{:});

XHD = X'*HD;
XHD = 0.5*(XHD + XHD');

% DX = D'*X;
% DX = 0.5*(DX+DX');
% 
% DG = D'*grad;
% DG = 0.5*(DG'+DG);


DXpG = D'*(grad - (2*beta)*X);
DXpG = 0.5*(DXpG+ DXpG');

output = HD - D*(Lambda)- X*(XHD + DXpG  );

end