function [output] = stop_proxorth(fun, opts, varargin)
%
% using multipliers correction methods to solve the following optimization problem
% over the Stiefel manifold
%
%       min      f(X)
%       s.t.     X'*X = I,  where X in R^{n*p}.
%
% ----------------------------------
%  Input:
%
%                       fun --- a matlab function for f(X)
%                                     call:  [fvalue, gradf] = fun(X, data1, data2);
%                                   fvalue:  function value f(X)
%                                    gradf:  gradient of f(X)
%                                     data:  extra data (can be more)
%                    
%                      opts --- option structure with fields:
%                                        X:  n-by-p initial point such that X'*X=I
%                                      dim:
%                                    maxit:  max iteration
%                             info_warning:
%                                proxparam:  proximal parameter
%                                 stepsize:  
%                                     info:  0(no print) o.w.(print)
%                                     xtol:  stop control for ||X_k - X_{k+1}||/sqr(n)
%                                     gtol:  stop control for ||kkt||/||kkt0||
%                                   solver:  1(gradient reflection method) 2(projection method) 3(QR retraction)
%                                      sym:  symmetrization times

%                  varargin --- extra data or other information
%
%
%  Output:
%                    output --- output information
%                                        X:  solution
%                                      kkt:  ||kkt|| (first-order optimality condition)
%                                     fval:  function value of solution
%                                     feaX:  ||I - X'X||_F (feasiblity violation)
%                                     xerr:  ||X_k - X_{k+1}||/sqr(n)
%                                     iter:  total iteration number
%                                    fvals:  function value history
%                                     kkts:  KKT violation history
%                                  message:  convergence message
%
%
%  Calling syntax:
%                     [output] = ProxOrth(@fun, opts, data1, data2);
%
%
% --------------------------------------------------------------------
%  Reference:
%    L. Wang, B. Gao, and X. Liu
%    Multipliers Correction Methods for Optimization Problems over the Stiefel Manifold
%    https://arxiv.org/abs/2011.14781
%   
% ----------------------------------
%  Author: Lei Wang, Bin Gao, and Xin Liu (ICMSEC, AMSS, CAS)
%                wlkings@lsec.cc.ac.cn
%                 gaobin@lsec.cc.ac.cn
%                 liuxin@lsec.cc.ac.cn
%  Version: 2.0 --- 2021/7/16
% --------------------------------------------------------------------

%% default setting
% if nargin < 3
%     opts = [];
% end

%% Warnings 
if ~isfield(opts, 'info_warning')
    info_warning = 0;
else
    info_warning = opts.info_warning;
end


%% Specify the initial point
if isfield(opts, 'X') && (~isempty(opts.X))
    [n, p] = size(opts.X);
    X = opts.X;
elseif isfield(opts, 'dim')
    n = opts.dim(1);
    p = opts.dim(2); % �????dim??size
    if n < p
        error('Column size should be no less than the size of row. Please check the dimensions of the opts.dim.') % ??�????, n\& p???��???�??
    end
    [X, ~] = qr(randn(n, p), 0);
else
    error('No initial point, and the size of initial point is not specified.')
end


%% Specify the proximal parameter according to ||\nabla f(X_0)||_F
[fvalue, gradf] = feval(fun, X, varargin{:});
s_upper = norm(gradf, 'fro');

if ~isfield(opts, 'proxparam')
    gamma = max(min(1e-3 * s_upper / p, 1e10), 1e-10); % set lower-bound for beta
else
    gamma = opts.proxparam;  % warning for too large/small beta
    if info_warning
        if gamma > 1e10
            warning('The proximal parameter is too large. Please scale the problem or specify opts.proxparam.');
        elseif gamma < 1e-10
            warning('The proximal parameter is too small. Please scale the problem or specify opts.proxparam.');
        end
    end
end

gamma = 1e-3;

%% Riemannian gradient
Lambda_0= X' * gradf;
Lambda = 0.5 * (Lambda_0 + Lambda_0');
rgradf = gradf - X * Lambda;


%% Specify the options for stepsize
if isfield(opts, 'stepsize') 
    if ~isfield(opts.stepsize, 'type')
        stepsize_type = 0;
        if info_warning
            warning('No input stepsize type, choose alternating BB stepsize.')
        end
    else
        if strcmp(opts.stepsize.type, 'ABB')
            stepsize_type = 0;   
        elseif strcmp(opts.stepsize.type, 'BB1')
            stepsize_type = 1;
        elseif strcmp(opts.stepsize.type, 'BB2')
            stepsize_type = 2;
        elseif strcmp(opts.stepsize.type, 'SD') % initial stepsize by difference
            stepsize_type = 3;
        else
            stepsize_type = 0; 
            if info_warning
                warning('Invaild stepsize type, choose ABB instead.')
            end
        end
    end
    
    if ~isfield(opts.stepsize, 'max_stepsize')
        max_stepsize = 1000;
    else
        max_stepsize = opts.stepsize.max_stepsize;
    end

    if ~isfield(opts.stepsize, 'min_stepsize')
        min_stepsize = 0;
    else
        min_stepsize = opts.stepsize.min_stepsize;
    end
        
    if ~isfield(opts.stepsize, 'init_stepsize')
        init_stepsize = 1 / (norm(X, 'fro') / 2 * norm(rgradf, 'fro') + p * gamma);
    else
        init_stepsize = opts.stepsize.init_stepsize;
    end
else
    if info_warning 
        warning('No input stepsize.')
    end
    stepsize_type = 0;
    max_stepsize = 1000;
    min_stepsize = 0;
    init_stepsize = 1 / (norm(X, 'fro') / 2 * norm(rgradf, 'fro') + p * gamma);
end

%% Options for line search
if ~isfield(opts, 'linesearch')
    linesearch = 0;
else
    linesearch = opts.linesearch;
end


%% Maximum iterations
if ~isfield(opts, 'maxit')
    maxlength = 3000;
else
    maxlength = opts.maxit;
    if maxlength < 0 || maxlength > 10000
        maxlength = 3000;
    end
end


%% Options for tolerence
if ~isfield(opts, 'gtol') % add gtol
    gtol = 1e-3;
else
    gtol = opts.gtol;
    if gtol < 0 || gtol > 1
        gtol = 1e-3;
        if info_warning
            warning('Invaild value of gtol, choose 1e-3 instead.')
        end
    end

end
if ~isfield(opts, 'xtol') % add xtol
    xtol = 1e-6;
else
    xtol = opts.xtol;
    if xtol < 0 || xtol > 1
        xtol = 1e-6;
        if info_warning
            warning('Invaild value of xtol, choose 1e-6 instead.')
        end
    end

end


%% Options for printed infomation
if ~isfield(opts, 'info')
    local_info = 2;
else
    local_info = opts.info;
end

%% Specify the solver
if isfield(opts, 'solver')
    if strcmp(opts.solver, 'GR')
        solver = 1;   
    elseif strcmp(opts.solver, 'GP')
        solver = 2;
    elseif strcmp(opts.solver, 'QR')
        solver = 3;
    else
        solver = 2; 
        if info_warning
            warning('Invaild solver type, choose GP instead.')
        end
    end
else
    if info_warning 
        warning('No input solver, choose GP by default.')
    end
    solver = 2;
end

%% Options for proximal correction step
if isfield(opts, 'sym')
    if opts.sym < 0
        if opts.sym ~= -1
            opts.sym = -1;
        end
    elseif opts.sym ~= fix(opts.sym)
        opts.sym = -1;
    end
else
    opts.sym = -1;
end

if norm(Lambda_0 - Lambda_0', 'fro') <= 1e-8
    opts.sym = 0;
end
sym = opts.sym;

%% ---------------------------------------------------------------

% Initialization
XX = X' * X;
feas_init = norm(XX - speye(p), 'fro');
kkts_init =  norm(rgradf, 'fro');
fval_init = fvalue;


violation_KKT = [kkts_init];
violation_Constraint = [feas_init];
fval = [fval_init];

line_search_record = [0];

time_record = [0];
t_start = tic;

if local_info
    switch solver
        case 1
            fprintf('------------------ ProxOrth with gradient reflection start ------------------\n');
        case 2
            fprintf('------------------ ProxOrth with gradient projection start ------------------\n');
        case 3
            fprintf('------------------ Riemannian Opt with QR retraction start ------------------\n');
    end
    fprintf('%4s |  %10s  |  %8s   |  %10s   |  %10s   |     %8s   \n', 'Iter ', 'F(X) ', 'KKT ', 'Xerr ', 'Fea ', 'stepsize');
    fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n', 0, fval_init, kkts_init, 0, feas_init, 0);
end

% ensure X is orthogonal
if norm(X' * X - speye(p), 'fro') > 1e-13
    [X, ~] = qr(X, 0);
end

% initial solver
switch solver
    case 1; mainsolver = @gradient_reflection;
    case 2; mainsolver = @gradient_projection;
    case 3; mainsolver = @QR_retraction;
end

%% ---------------------------------------------------------------
% main iteration
for iter = 2 : 1 : maxlength

    X_pre = X;     
    gradf_pre = gradf;    
    rgradf_pre = rgradf;   

    % Computing stepsize
    if stepsize_type ~= 3
        if iter < 3 
            stepsize = init_stepsize;
        else
            stepsize_flag = mod(iter, 2);
            if stepsize_type == 2
                stepsize_flag = 0;
            elseif stepsize_type == 1
                stepsize_flag = 1;
            end
            if stepsize_flag == 0
                stepsize =  abs(norm(D, 'fro')^2 / sum(sum(Grad_Y .* D)));
            else
                stepsize = abs(sum(sum(Grad_Y .* D)) / norm(Grad_Y, 'fro')^2);
            end
        end
    else
        if iter < 3
            stepsize = init_stepsize;
        else
            stepsize = 2 * abs(fval(end - 1) - fval(end)) / violation_KKT(end)^2;
        end
    end
    stepsize = max(min_stepsize, min(abs(stepsize), max_stepsize));

    
    % ---------- PART I:    gradient step ----------
    %  Riemmanian or Euclidean gradient
    if solver == 3
        Grad = rgradf_pre;
    else
        Grad = gradf_pre;
    end
    
    
    if linesearch == 0
        % gradient step
        V = X - stepsize * Grad;
        X = mainsolver(X, V);
        ls_iter = 0;
    else
        local_fval_grad = @(ZZ) feval(fun, ZZ, varargin{:});
        if stepsize_type ~= 3
            [X, ls_iter] = ProxOrth_lineasarch(X, Grad, stepsize, local_fval_grad, mainsolver, fval);
        else
            [X, ls_iter] = ProxOrth_lineasarch_sd(X, Grad, stepsize, local_fval_grad, mainsolver, fval);
        end
    end
    line_search_record = [line_search_record, ls_iter];
    
    % ---------- PART II: correction step ----------
    if norm(X' * X - speye(p), 'fro') > 1e-12
        [X, ~] = qr(X, 0); 
    end
            
    if opts.sym == -1
        sym = 2 * ceil(0.5 * sqrt(iter)) - 1;
    end
    
    if sym > 0
        if solver ~= 3
            for symtime = 1 : sym
                [~, gradf] = feval(fun, X, varargin{:});
                Z = X' * gradf - gamma * speye(p);
                [tu, ~, tv] = svd(Z, 0);
                X = -X * (tu * tv');
            end
        end
    end

    
    % ------------ evaluate error ------------
    [fvalue, gradf] = feval(fun, X, varargin{:});
    Lambda = X' * gradf;
    Lambda = (Lambda + Lambda') / 2;
    rgradf = gradf - X * Lambda;

    D = X - X_pre;
    Grad_Y = rgradf - rgradf_pre;
    
    violation_KKT = [violation_KKT, norm(rgradf, 'fro')];
    violation_Constraint = [violation_Constraint, norm(X' * X - speye(p), 'fro')];
    time_record = [time_record, toc(t_start)];
    fval = [fval, fvalue];

    Xerr = norm(D, 'fro');

    if local_info && mod(iter, 20) == 0 
        fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n', iter, fvalue, violation_KKT(iter), Xerr, violation_Constraint(iter), stepsize);
    end
    
    if violation_KKT(iter) < gtol * kkts_init 
        if local_info
            fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n', iter, fvalue, violation_KKT(iter), Xerr, violation_Constraint(iter), stepsize);
            fprintf('The norm of Riemannian gradient is less than tolerence.\n')
        end
        break;
    end
    
    if Xerr < xtol 
        if local_info
            fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n', iter, fvalue, violation_KKT(iter), Xerr, violation_Constraint(iter), stepsize);
            fprintf('The norm of X_{k} - X_{k-1} is less than tolerence.\n')
        end
        
        break;
    end

    if violation_KKT(iter) > 1e12
        warning('The algoorithm may diverge.\n');
        break;
    end
    
end

if iter >= maxlength
    warning('The algorithm reaches maximum iteration.\n');
end


% Output
output.X = X;
output.iter = iter;

output.kkt = violation_KKT(end);
output.fval = fvalue;
output.fea = violation_Constraint(end);

output.kkts = violation_KKT;
output.fvals = fval;
output.feas = violation_Constraint;
output.times = time_record;

output.line_search_iters = line_search_record;



%% ---------------------------------------------------------------
% nest-function

% gradient reflection step
    function X = gradient_reflection(X, V)
        VV = V' * V;
        VX = V' * X;
        TVX = VV \ VX;
        X = -X + V * (2 * TVX);
    end

% projection step
    function X = gradient_projection(~, V)
        % approach 1
        VV = V' * V;
        [Q, d] = eig(VV);
        DD = sqrt(d) \ eye(p);
        X = V * (Q * DD * Q');
        
        % approach 2
        %  [UX,~,VX] = svd(V,0);
        %   X = UX*VX';
    end

% QR retraction
    function X = QR_retraction(~, V)
        % approach 1
        VV = V' * V;
        L = chol(VV, 'lower');
        X = V * inv(L)';
        
        % approach 2
        %  [X, ~] = qr(V, 0);
    end

end


function [out, iter] = ProxOrth_lineasarch(X, Grad, stepsize, fval_grad, mainsolver, fval_records)

    if isempty(fval_records)
        V = X - stepsize * Grad;
        X = mainsolver(X, V);
        iter = 0;
    else
        stepsize_local = stepsize;
        [n, p] = size(X);
        for jj = 0: 1 : 2
            V = X - stepsize_local * Grad;
            Y = mainsolver(X, V);
            [fval, ~] = fval_grad(Y);
            if fval <= max(fval_records(max(1, end - 50) : end))
                break
            else
                stepsize_local = stepsize_local * 0.5;
            end
        end
        iter = jj;
        out = Y;
    end

end


function [out, iter] = ProxOrth_lineasarch_sd(X, Grad, stepsize, fval_grad, mainsolver, fval_records)
    
    if isempty(fval_records)
        V = X - stepsize * Grad;
        X = mainsolver(X, V);
        iter = 0;
    else
        stepsize_local = stepsize;
        [n, p] = size(X);
        for jj = 0: 1 : 5
            V = X - stepsize_local * Grad;
            Y = mainsolver(X, V);
            [fval, ~] = fval_grad(Y);
            if fval <= max(fval_records(end))
                break
            else
                stepsize_local = stepsize_local * 0.5;
            end
        end
        iter = jj;
        out = Y;
    end

end