function [output] = stop_slpg(fun,opts,varargin)
% Sequential linearized proximal gradient (SLPG) metnod for nonsmooth optimization over the Stiefel manifold
% function [output] = stop_slpg(fun,opts)
% function [output] = stop_slpg(fun,opts,varargin)
% Apply SLPG to the following nonsmooth optimization problems on the Stiefel manifold,

% min f(X) + r(x) 
% s.t. X^TX = I_p.

% Here f(X) is differentiable, and r(X) is convex but possible nonsmooth. 
% Besides, the proximal mapping of r(X), i.e. 
% min_{D \in R^{n\times p}}  ||D - X||_F^2 + r(X) 
% is easy to solve. 


% The descriptions for stop_slpg solver could also be found at https://stmopt.gitee.io/algorithm_description/SLPG_code.html

% The SLPG solver starts at opts.X if it is provided (otherwise, generates opts.X whose size is specified in
% opts.dim on the Stiefel manifold). 
% The fun in SLPG solver is called by
% [fval, grad] = fun(X) or [fval, grad] = fun(X, varargin)

% The opts in stop_slpg solver is a matlab structure for overwritting the default values. 
% In opts structure, either opts.X or opts.dim should be specified. 
% Any other option have a default value and is hence optional. To assign the value to the opts structure,
% one needs to pass an options structure with a field options.optionname.


% *  opts.X
%   * The initial point for the solver.  opts.X  should be a matrix with $n$ rows and $p$ columns, where $n\geq p$. We recommend to choose a feasible  opts.X  in practice. 
%   * When  opts.X  is empty, the option  opts.dim  should be specified, and the solver will randomly generate a feasible initial point by  [X, ~] = qr(randn(n,p),0) ; otherwise, the solver will raise an error. 
% *  opts.gamma = 0 
%   * the parameter before the regularization term. Its default value is $0$, but we strongly recommend to specify it in  opts .
% *  opts.prox 
%   *  [Y] = opts.prox(X, G, gamma, eta) : 
%   * Returns the solution of  $Y = \arg \min_{D \in \mathbb{R}^{n\times p}} <X, G> + \gamma r(X)+ \frac{1}{2\eta}||D-X||_2$
%   * The default value is the proximal mapping for $r(X) = ||X||_1$.

% *  opts.stepsize 
%   *  Stepsize arguments for SLPG
%   *   opts.stepsize.type = 'BB1' :  
%      *  Currently,  stop_slpg only supports BB1 stepsize. 
%   *   opts.stepsize.max_stepsize = 1000 : Upper-bound for stepsize
%   *   opts.stepsize.min_stepsize = 0 : Lower-bound for stepsize
%   *   opts.stepsize.init_stepsize : Initial stepsize
% *  opts.gtol = 1e-8 
%   * Tolerance for the kkt condition
%   * SLPG terminates when $\frac{1}{\eta_{k-1}}\norm{X_k - X_{k-1}}_F \leq opts.gtol$
% *  opts.maxit = 1000 
%   * Maximum iteration
%   * SLPG terminates when the iteration exceeds  maxit 
% *  opts.postorth = 1 
%   * Option for orthonormalization as postprocess:
%   * The postprocess improves the feasibility but requires additional computational costs. We recommend to set   opts.postorth = 1  when high precision in feasibility is required. 
% *  opts.info = 0 
%   * Option for display during the execution
%   * No print or display when  opts.info = 0 
% *  opts.info_warning = 0 
%   * Options for displaying warnings in the solver. 


% All the outputs of SLPG_smooth is wrapped in the output structure. 

% *  output : output information
%   *  output.X 
%     * Final solution
%   *  output.iter 
%     * Total iterations taken by PenCF
%   *  output.fval 
%     * Final function value
%   *  output.kkt 
%     * Final substationarity evaluated by $||\nabla f(X_k) - X_k \Lambda(X_k)||_F$
%   *  output.fea 
%     * Final feasibility evaluated by $||X_k^\top X_k - I_p||_F$ 
%   *  output.fvals 
%     * Records of function values in each iteration
%   *  output.kkts 
%     * Record of the substationarities in each iteration
%   *  output.feas 
%     * Record of the feasibility in each iteration

% The implementation of this solver is based on the following paper:

% N. Xiao, X. Liu and Y. Yuan, A Penalty-free Infeasible Approach for a Class of Nonsmooth Opimtization Problems over the Stiefel Manifold
% https://arxiv.org/abs/2103.03514v3
% 


% Original author: Nachuan Xiao, Xin Liu, Ya-xiang Yuan, 2021/10
% Contributors: Nachuan Xiao, Xin Liu, Ya-xiang Yuan

% Change log:




if isfield(opts, 'X') && (~isempty(opts.X))
    [n,p] = size(opts.X);
    X = opts.X;
elseif isfield(opts, 'dim')
    n = opts.dim(1);
    p = opts.dim(2);
    if n < p
        error('n is smaller than p. Please check the dimensions of the opts.dim.')
    end
    [X, ~] = qr(randn(n,p),0);
else
    error('No initial point, and the size of initial point is not specified')
end
    
% X = Xinit;

if ~isfield(opts, 'maxit')
    maxlength = 3000;
else
    maxlength = opts.maxit;
end


if ~isfield(opts, 'info_warning')
    info_warning = 0;
else
    info_warning = opts.info_warning;
end
Phi =  @(M) 0.5*(M+M');


[n,p] = size(X);
[fval_init, gradf] = feval(fun,X,varargin{:});






if isfield(opts,'stepsize') 
    if ~isfield(opts.stepsize,'type')
        stepsize_type = 0;
        if info_warning
            warning('No input stepsize type, choose alternating BB stepsize. ')
        end
    else
        if strcmp(opts.stepsize.type, 'ABB')
            stepsize_type = 0;   
        elseif strcmp(opts.stepsize.type, 'BB1')
            stepsize_type = 1;
        elseif strcmp(opts.stepsize.type, 'BB2')
            stepsize_type = 2;
        elseif strcmp(opts.stepsize.type, 'FIX')
            stepsize_type = 3;
        else
            stepsize_type = 0; 
            if info_warning
                warning('Invaild stepsize type, choose ABB instead. ')
            end
        end
        
    end
    
    if ~isfield(opts.stepsize,'max_stepsize')
        max_stepsize = 100;
    else
        max_stepsize = opts.stepsize.max_stepsize;
    end

    if ~isfield(opts.stepsize,'min_stepsize')
        min_stepsize = 0;
    else
        min_stepsize = opts.stepsize.min_stepsize;
    end
    
    
    if ~isfield(opts.stepsize,'init_stepsize')
        init_stepsize = 0.5/norm(gradf,2);
    else
        init_stepsize = opts.stepsize.init_stepsize;
    end
else
    if info_warning 
        warning('No input stepsize. ')
    end
    stepsize_type = 0;
    max_stepsize = 1000;
    min_stepsize = 0;
    init_stepsize = 0.5/norm(gradf,2);
end


%% Options for tolerence

if ~isfield(opts,'gtol')
    gtol = 1e-8;
else
    gtol = opts.gtol;
end

%% Options for printed infomation
if ~isfield(opts,'info')
    local_info = 2;
else
    local_info = opts.info;
end



%% Options for proximal operator
if ~isfield(opts,'prox')
    my_prox = @my_prox_l1;
else
    my_prox = opts.prox;
end


%% Options for post-process
if ~isfield(opts,'postorth')
    postprocess = 1;
else
    postprocess = opts.postorth;
end




if ~isfield(opts, 'gamma')
    
    warning('Not specify coefficient for regularization term.Please specify one.\n')
    gamma = 0;
else
    gamma = opts.gamma;
end


norm_hessian_esti = norm(gradf,'fro');
violation_KKT = zeros(1,maxlength);
violation_Constraint = zeros(1,maxlength);
inner_iter = zeros(1,maxlength);
fval = zeros(1,maxlength);
stepsize_record = zeros(1,maxlength);
record_sub_feas = zeros(1,maxlength);

line_search_record = zeros(1,maxlength);

%Lambda_f = X'*gradf;
Lambdar = zeros(p,p); % initialization
%Lambdar =gamma* Phi(X'*sign(X)); % initialization when X starts nearby a stationary point

Lambdaf = Phi(X'*gradf);
Gradf = gradf - X*Lambdaf;
Grad = gradf - X*(Lambdar+ Lambdaf);
feas_init = norm(X'*X - eye(p),'fro');

if local_info
    fprintf('%4s | %12s | %10s | %10s | %8s | %8s\n', 'Iter ', 'F(X) ', 'KKT ', 'Xerr ', 'Fea ', 'stepsize');
    fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',0, fval_init, 0, 0, feas_init, 0);
end
t_start = tic;

time_record = zeros(1,maxlength);

for jj = 1:1:maxlength
    
    if jj <  2 
        stepsize = init_stepsize;% choose fixed stepeize in the early stage to make the algorithm more stable
    else
        
        stepsize = norm(DX,'fro')^2 / sum(sum(DX.*Grad_Y)); % choose BB1 stepsize
       
    end


    stepsize = max(min_stepsize,  min(abs(stepsize), max_stepsize));
    
    
    
    X_p = X;
    
    
    
    X = my_prox(X,Grad,gamma,stepsize); % update X in the null space
    
    
    
    
    
    %X = X./sqrt(sum(X.*X,1));
    XX = X'*X;
    
    if norm(XX - eye(p),'fro') < 0.3 
    X = X- 0.5*X*(XX-eye(p));  % compute the soft retraction
    else
    
    X = 2*X/(eye(p) + XX);    % more stable but less efficient
    end
    
    [fvalue,gradf] = feval(fun,X,varargin{:});  % compute the gradient
    fval(jj) = fvalue;

    num_linsea = 0;
    

   
    
    stepsize_record(jj) = stepsize;     % record for stepsize
    
    line_search_record(jj) = num_linsea;
    
    
    
    
    XX = X'*X;


    
    
    DX = X-X_p;
    
    
    
    Lambdaf = Phi(X'*gradf);  % compute the multiplier for smooth part
    
    
    Grad_p = Gradf;  % store the previous gradient for the smooth part
    Gradf = gradf+ X*( - Lambdaf); % compute the new approximated gradient
    Lambda_temp = Lambdar; % warm-start for Lambdar,
    
    
    
    %normXX = norm(XX - eye(p),'fro');
    for jr = 1:1:10 % recommend to be larger than 10
        
        if jj > 2
            try_stepsize = max(min(mean(stepsize_record(max(1,jj-50):jj-1) ), 10000/norm_hessian_esti), 0.01/norm_hessian_esti) ;
        else
            try_stepsize  = max(min(stepsize, 100/norm_hessian_esti), 0.01/norm_hessian_esti) ;
        end
        X_try = my_prox(X,Gradf+ X*( -Lambda_temp),gamma,try_stepsize);
       % DDD =  -0.4/try_stepsize*((2*X_try - 1*X_try_p - X)); % PDA by Chambolle-Pock
        DDD =  -1/try_stepsize*((X_try - X)) ; % Arrow-Hurwicz method
        
        Lambda_inc = Phi(X'*DDD);
        
        Lambda_temp = Lambda_temp + Lambda_inc;
        
        if jj > 1 && norm(Lambda_inc,'fro') < 1e-2* mean(violation_KKT( max(1,jj-20):jj-1 ) ) % stopping criteria for subproblem
            break
        end
        

    end
    
    
    inner_iter(jj) = jr; % record the inner iteration
    record_sub_feas(jj) = norm(Lambda_inc,'fro');
    Lambdar = Lambda_temp;
    Lambda = Lambdaf+ Lambdar; % true multiplier
    
    
    Grad = gradf+ X*( - Lambda);
    Grad_Y = Gradf - Grad_p - DX*( Lambdar  );

    violation_KKT(jj) = norm(DX,'fro')/stepsize;
    violation_Constraint(jj) = norm(XX - eye(p),'fro');
    fval(jj) = fvalue;
    time_record(jj) = toc(t_start);
    if local_info && mod(jj,20) == 0 
        fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e \n',jj, fvalue, violation_KKT(jj), norm(DX,'fro'), violation_Constraint(jj), stepsize);
    end
    
    
    if violation_KKT(jj) < gtol
        
        break;
    end
    
    if violation_KKT(jj) > 1e16
        break;
    end
end


if local_info 
    fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',jj, fvalue, violation_KKT(jj), norm(DX,'fro'), violation_Constraint(jj), stepsize);
end

violation_Constraint(1) = 1;
record_sub_feas(1) = 1;
output.vio = violation_KKT(1:jj);

output.iter = jj;

if postprocess
    if local_info 
        fprintf('Post-processing \n')
    end
    [X,~] = qr(X,0);             % orthonormalization in the final stage
end

X(find(abs(X) < 1e-5)) = 0;  % thresholding



[fvalue,gradf] = feval(fun,X,varargin{:});
fval(jj) = fvalue;


output.X = X;
output.iter = jj;

output.kkt = violation_KKT(jj);
output.fval = fvalue;
output.fea = violation_Constraint(jj);


output.fvals = fval(1:jj);
output.feas = violation_Constraint(1:jj);
output.kkts = violation_KKT(1:jj);

output.inner_iters = inner_iter(1:jj);
output.stepsizes = stepsize_record(1:jj);
output.feas_subs = record_sub_feas(1:jj);
output.times = time_record(1:jj);

end


function [Y] =  my_prox_l1(X,D,gamma,eta)
% solves the problem min X^TD + gamma*||Y||_1 + 1/(2*eta)*||Y-X||_2

%[n,p] = size(X);
Xt = X - eta*D;
Y = min(0, Xt + gamma*eta) + max(0, Xt - gamma*eta);
end
