function [output] = stop_slpg_l21(fun,opts,varargin)

% Sequential linearized proximal gradient (SLPG) metnod for L21-norm optimization over the Stiefel manifold
% function [output] = stop_slpg_l21(fun,opts)
% function [output] = stop_slpg_l21(fun,opts,varargin)
% Apply SLPG to the following nonsmooth optimization problems on the Stiefel manifold,

% min f(X) + r(x) 
% s.t. X^TX = I_p.

% Here f(X) is differentiable, 
% and r(X) = \gamma \sum_{i = 1}^n  || X(i,:) ||_2
% is a L21 norm regularizer. 


% The descriptions for stop_slpg_l21 solver could also be found at https://stop.gitee.io/algorithm_description/SLPG_l21_code.html

% The stop_slpg_l21 solver starts at opts.X if it is provided (otherwise, generates opts.X whose size is specified in
% opts.dim on the Stiefel manifold). 
% The fun in stop_slpg_l21 solver is called by
% [fval, grad] = fun(X) or [fval, grad] = fun(X, varargin)

% The opts in this solver is a matlab structure for overwritting the default values. 
% In opts structure, either opts.X or opts.dim should be specified. 
% Any other option have a default value and is hence optional. To assign the value to the opts structure,
% one needs to pass an options structure with a field options.optionname.



% * opts.X
%   * The initial point for the solver. opts.X should be a matrix with $n$ rows and $p$ columns, where $n\geq p$. We recommend to choose a feasible opts.X in practice. 
%   * When opts.X is empty, the option opts.dim should be specified, and the solver will randomly generate a feasible initial point by [X, ~] = qr(randn(n,p),0); otherwise, the solver will raise an error. 
% * opts.gamma = 0
%   * the parameter before the regularization term. Its default value is $0$, but we strongly recommend to specify it in opts.
% * opts.stepsize
%   *  Stepsize arguments for the solver
%   *  opts.stepsize.type = 0:  
%      *  opts.stepsize.type = 'ABB': Alternating BB stepsize
%      *  opts.stepsize.type = 'BB1': BB1 stepsize
%      *  opts.stepsize.type = 'BB2': BB2 stepsize
%      *  opts.stepsize.type = 'FIX': fixed stepsize
%   *  opts.stepsize.max_stepsize = 1000: Upper-bound for stepsize
%   *  opts.stepsize.min_stepsize = 0: Lower-bound for stepsize
%   *  opts.stepsize.init_stepsize: Initial stepsize
% * opts.gtol = 1e-8
%   * Tolerance for the kkt condition
%   * Due to the close-form updating scheme, the solver terminates when $||\nabla f(X_k) - X_k \Lambda(X_k) ||_F \leq \mathrm{opts.gtol}$.
% * opts.maxit = 1000
%   * Maximum iteration
%   * solver terminates when the iteration exceeds maxit
% * opts.postorth = 1
%   * Option for orthonormalization as postprocess:
%   * The postprocess improves the feasibility but requires additional computational costs. We recommend to set  opts.postorth = 1 when high precision in feasibility is required. 
% * opts.info = 0
%   * Option for display during the execution
%   * No print or display when opts.info = 0
% * opts.info_warning = 0
%   * Options for displaying warnings in the solver. 


% All the outputs of SLPG_smooth is wrapped in the output structure. 

% * output: output information
%   * output.X
%     * Final solution
%   * output.iter
%     * Total iterations taken by the solver
%   * output.fval
%     * Final function value
%   * output.kkt
%     * Final substationarity evaluated by $||\nabla f(X_k) - X_k \Lambda(X_k)||_F$
%   * output.fea
%     * Final feasibility evaluated by $||X_k^\top X_k - I_p||_F$ 
%   * output.fvals
%     * Records of function values in each iteration
%   * output.kkts
%     * Record of the substationarities in each iteration
%   * output.feas
%     * Record of the feasibility in each iteration



% The implementation of this solver is based on the following papers:
%
% N. Xiao, X. Liu and Y. Yuan, Exact Penalty Function for L21 Norm Minimization over the Stiefel Manifold, SIAM Journal on Optimization (accepted).
% http://www.optimization-online.org/DB_HTML/2020/07/7908.html

% and

% N. Xiao, X. Liu and Y. Yuan, A Penalty-free Infeasible Approach for a Class of Nonsmooth Opimtization Problems over the Stiefel Manifold
% https://arxiv.org/abs/2103.03514v3
% 


% Original author: Nachuan Xiao, Xin Liu, Ya-xiang Yuan, 2021/10
% Contributors: Nachuan Xiao, Xin Liu, Ya-xiang Yuan

% Change log:



%% Specify the initial point
if isfield(opts, 'X') && (~isempty(opts.X))
    [n,p] = size(opts.X);
    X = opts.X;
elseif isfield(opts, 'dim')
    n = opts.dim(1);
    p = opts.dim(2);
    if n < p
        error('n is smaller than p. Please check the dimensions of the opts.dim.')
    end
    [X, ~] = qr(randn(n,p),0);
else
    error('No initial point, and the size of initial point is not specified')
end
    
% X = Xinit;

if ~isfield(opts, 'maxit')
    maxlength = 3000;
else
    maxlength = opts.maxit;
end


if ~isfield(opts, 'info_warning')
    info_warning = 0;
else
    info_warning = opts.info_warning;
end

%% Specify the penalty parameter according to ||\nabla f(X_0)||_F
[fvalue, gradf] = feval(fun,X,varargin{:});
s_upper = norm(gradf,'fro');

if ~isfield(opts, 'penalparam')
     beta = 0.01*s_upper;
else
    beta = opts.penalparam;
end




XX = X'*X;

Lambdah = work_compute_Lambdah(X);
Lambda = X'*gradf;
Lambda = 0.5*(Lambda+Lambda') + Lambdah;
Grad = gradf - X*(Lambda  - beta * ( XX- eye(p))  )  ; % approximated gradient



if ~isfield(opts, 'gamma')
     gamma = 0;
     warning('gamma is not specified in options. Please check. ')
else
    gamma = opts.gamma;
end







%% Specify the options for stepsize
if isfield(opts,'stepsize') 
    if ~isfield(opts.stepsize,'type')
        stepsize_type = 0;
        if info_warning
            warning('No input stepsize type, choose alternating BB stepsize. ')
        end
    else
        if strcmp(opts.stepsize.type, 'ABB')
            stepsize_type = 0;   
        elseif strcmp(opts.stepsize.type, 'BB1')
            stepsize_type = 1;
        elseif strcmp(opts.stepsize.type, 'BB2')
            stepsize_type = 2;
        elseif strcmp(opts.stepsize.type, 'FIX')
            stepsize_type = 3;
        else
            stepsize_type = 0; 
            if info_warning
                warning('Invaild stepsize type, choose ABB instead. ')
            end
        end
        
    end
    
    if ~isfield(opts.stepsize,'max_stepsize')
        max_stepsize = 1000;
    else
        max_stepsize = opts.stepsize.max_stepsize;
    end

    if ~isfield(opts.stepsize,'min_stepsize')
        min_stepsize = 0;
    else
        min_stepsize = opts.stepsize.min_stepsize;
    end
    
    
    if ~isfield(opts.stepsize,'init_stepsize')
        init_stepsize =1/( norm(X,'fro')/2*norm(Grad,'fro') + p* beta);
    else
        init_stepsize = opts.stepsize.init_stepsize;
    end
else
    if info_warning 
        warning('No input stepsize. ')
    end
    stepsize_type = 0;
    max_stepsize = 1000;
    min_stepsize = 0;
    init_stepsize = 1/( norm(X,'fro')/2*norm(Grad,'fro') + p* beta);
end

%% Options for tolerence

if ~isfield(opts,'gtol')
    gtol = 1e-8;
else
    gtol = opts.gtol;
end

%% Options for printed infomation
if ~isfield(opts,'info')
    local_info = 2;
else
    local_info = opts.info;
end


%% Options for post-process
if ~isfield(opts,'postorth')
    postprocess = 1;
else
    postprocess = opts.postorth;
end

%% Options for line search

if ~isfield(opts,'linesearch')
    linesearch = 1;
else
    linesearch = opts.linesearch;
end

%%
if ~isfield(opts,'warmup_iter')
    warmup_iter = 2;
else
    warmup_iter = opts.warmup_iter;
end


%% Initialization

violation_KKT = zeros(1,maxlength+1);
violation_Constraint = zeros(1,maxlength+1);
fval = [];

line_search_record = [];

time_record = zeros(1,maxlength+1);
t_start = tic;

feas_init = norm(XX- eye(p),'fro');
kkts_init =  1;
fval_init = fvalue;

if local_info
    fprintf('%4s | %12s | %10s | %10s | %8s | %8s\n', 'Iter ', 'F(X) ', 'KKT ', 'Xerr ', 'Fea ', 'stepsize');
    fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',0, fval_init, kkts_init, 0, feas_init, 0);
end

%% Main loop
for jj = 1:1:maxlength
    X_pre = X;

    % Computing stepsize
    if jj < 2 || stepsize_type == 3
        stepsize = init_stepsize;
    else
        stepsize_flag = mod(jj,2);
        if stepsize_type == 2
            stepsize_flag = 0;
        elseif stepsize_type == 1
            stepsize_flag = 1;
        end
        if stepsize_flag == 0
            stepsize =  abs(norm(D,'fro')^2 /sum(sum(Grad_Y.*D)));
        else
            stepsize = abs( sum(sum(Grad_Y.*D))/ norm(Grad_Y,'fro')^2);
        end
    end
    stepsize = max(min_stepsize, min(abs(stepsize),max_stepsize));
    
    
    if 1 || linesearch == 0
        X = my_prox21(X,Grad,gamma,1*stepsize);
        ls_iter = 0;
    else
        local_fval_grad = @(ZZ) feval(fun,ZZ,varargin{:});
        [X, ls_iter] = PenCF_lineasarch(X, Grad, stepsize,local_fval_grad,fval, Lambda, XX, beta);
    end
    line_search_record = [line_search_record, ls_iter];
    
    
%     X = proj(X);
%     X = safe_proj(proj, X, stepsize*Lambda, violation_Constraint(max(jj-1,1)));
    
    for inner_proj_iter = 1:1:2
        XX = X'*X;
        norm_feas = norm(XX - eye(p), 'fro');
        if norm_feas <0.3
            X = X *(1.5 * eye(p) - 0.5 *XX);
            break
        elseif norm_feas < 3
            X = 2*X/(eye(p) + XX);
        else
            [X,~] = qr(X,0);
        end
    end
    XX = X'*X;
    
    
    D = X-X_pre;
    
    

    [fvalue,gradf] = feval(fun,X,varargin{:});
    
    Lambdah = work_compute_Lambdah(X);
    
    Lambda = X'*gradf;
    Lambda = 0.5*(Lambda+Lambda') + gamma * Lambdah;
    Grad_p = Grad;
    
    
    Grad = gradf - X*  (Lambda    );

    
    Grad_Y = Grad - Grad_p;

    violation_KKT(jj) =  norm((1/stepsize)*(X-X_pre),'fro');
    violation_Constraint(jj) = norm(XX- eye(p),'fro');
    time_record(jj) = toc(t_start);
    fval = [fval,fvalue - 0.5* sum(sum((XX- eye(p)).* Lambda)) + 0.25 * beta * norm(XX- eye(p),'fro')^2];
    
    if local_info && mod(jj,20) == 0 
        fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',jj, fvalue, violation_KKT(jj), norm(D,'fro'), violation_Constraint(jj), stepsize);
    end
    
    
    if violation_KKT(jj) < gtol * kkts_init 
        if local_info
            fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',jj, fvalue, violation_KKT(jj), norm(D,'fro'), violation_Constraint(jj), stepsize);
        end
        break;
    end

    if violation_KKT(jj) > 1e12
        jj = maxlength;
        break;
    end
end

if postprocess 
    if local_info 
        fprintf('Postprocess start \n')
    end
    X = post_orth(X);
    violation_Constraint(jj) = norm(X'*X- eye(p),'fro');
    
    if local_info
        fprintf('%04d \t %1.5e \t %3.2e \t %3.2e \t %3.2e \t %3.2e\n',jj, fvalue, violation_KKT(jj), 0, violation_Constraint(jj), 0);
    end
end

%% Output
output.X = X;
output.iter = jj;

output.kkt = violation_KKT(jj);
output.fval = fvalue;
output.fea = violation_Constraint(jj);

output.kkts = violation_KKT(1:jj);
output.fvals = fval(1:jj);
output.feas = violation_Constraint(1:jj);
output.times = time_record(1:jj);

output.line_search_iters = line_search_record;

end



function [out] =  my_prox21(X,D,gamma,eta)
% solves the problem min X^TD + gamma*||D||_{2,1} + 1/(2*eta)*||D-X||_2

[n,p] = size(X);
Xt = X - eta*D;
gamma = gamma*eta;
Xt_rho = sqrt(sum(Xt.*Xt,2));
% Xt_rho = vecnorm(Xt, 2,2);


Xt_rho_ref = 1 - gamma./(max(Xt_rho, gamma));

out = Xt_rho_ref.*Xt;
end


function Lambdah = work_compute_Lambdah(X)
%     [n,p] = size(X);
%     X_rho = sqrt(sum(X.*X,2));
%     index_X_rho = find(X_rho>=1e-14);
%     X_ref = zeros(n,1);
%     X_ref(index_X_rho) = 1./X_rho(index_X_rho);
%     Lambdah = X'*(X_ref.*X);

    XX_ref = bsxfun(@rdivide, X, (1e-14 + sqrt(sqrt(sum(X.*X,2))) ) );
    Lambdah = XX_ref' * XX_ref;
end


function [out] = post_orth(X)
[n,p] = size(X);

if p > 0.5 * n
    [U, ~, V] = svd(X,0);
    Y = U * V';
else 
    [V, D] = eig(X'*X);
    Y = X*(V*diag(sqrt(1./diag(D) ))*V');
end

[U, ~, V] = svd(X,0);
Z = U * V';



out = Y;
end